/*
Copyright (c) 2016-2017 Eugene Wissner

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Copyright: Eugene Wissner 2016-2017.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Eugene Wissner
 */
module dlib.async.event.epoll;

version (linux):

public import core.sys.linux.epoll;
import dlib.async.protocol;
import dlib.async.event.selector;
import dlib.async.loop;
import dlib.async.transport;
import dlib.async.watcher;
import dlib.memory;
import dlib.memory.mmappool;
import dlib.network.socket;
import core.stdc.errno;
import core.sys.posix.unistd;
import core.time;
import std.algorithm.comparison;

class EpollLoop : SelectorLoop
{
    protected int fd;
    private epoll_event[] events;

    /**
     * Initializes the loop.
     */
    this()
    {
        if ((fd = epoll_create1(EPOLL_CLOEXEC)) < 0)
        {
            throw MmapPool.instance.make!BadLoopException("epoll initialization failed");
        }
        super();
        events = MmapPool.instance.makeArray!epoll_event(maxEvents);
    }

    /**
     * Free loop internals.
     */
    ~this()
    {
        MmapPool.instance.dispose(events);
        close(fd);
    }

    /**
     * Should be called if the backend configuration changes.
     *
     * Params:
     *     watcher   = Watcher.
     *     oldEvents = The events were already set.
     *     events    = The events should be set.
     *
     * Returns: $(D_KEYWORD true) if the operation was successful.
     */
    protected override bool reify(ConnectionWatcher watcher, EventMask oldEvents, EventMask events)
    in
    {
        assert(watcher !is null);
    }
    body
    {
        int op = EPOLL_CTL_DEL;
        epoll_event ev;

        if (events == oldEvents)
        {
            return true;
        }
        if (events && oldEvents)
        {
            op = EPOLL_CTL_MOD;
        }
        else if (events && !oldEvents)
        {
            op = EPOLL_CTL_ADD;
        }

        ev.data.fd = watcher.socket.handle;
        ev.events = (events & (Event.read | Event.accept) ? EPOLLIN | EPOLLPRI : 0)
                  | (events & Event.write ? EPOLLOUT : 0)
                  | EPOLLET;

        return epoll_ctl(fd, op, watcher.socket.handle, &ev) == 0;
    }

    /**
     * Does the actual polling.
     */
    protected override void poll()
    {
        // Don't block
        immutable timeout = cast(immutable int) blockTime.total!"msecs";
        auto eventCount = epoll_wait(fd, events.ptr, maxEvents, timeout);

        if (eventCount < 0)
        {
            if (errno != EINTR)
            {
                throw defaultAllocator.make!BadLoopException();
            }
            return;
        }

        for (auto i = 0; i < eventCount; ++i)
        {
            auto io = cast(IOWatcher) connections[events[i].data.fd];

            if (io is null)
            {
                acceptConnections(connections[events[i].data.fd]);
            }
            else if (events[i].events & EPOLLERR)
            {
                kill(io, null);
                continue;
            }
            else if (events[i].events & (EPOLLIN | EPOLLPRI | EPOLLHUP))
            {
                auto transport = cast(SelectorStreamTransport) io.transport;
                assert(transport !is null);

                SocketException exception;
                try
                {
                    ptrdiff_t received;
                    do
                    {
                        received = transport.socket.receive(io.output[]);
                        io.output += received;
                    }
                    while (received);
                }
                catch (SocketException e)
                {
                    exception = e;
                }
                if (transport.socket.disconnected)
                {
                    kill(io, exception);
                    continue;
                }
                else if (io.output.length)
                {
                    swapPendings.insertBack(io);
                }
            }
            if (events[i].events & EPOLLOUT)
            {
                auto transport = cast(SelectorStreamTransport) io.transport;
                assert(transport !is null);

                transport.writeReady = true;
                if (transport.input.length)
                {
                    feed(transport);
                }
            }
        }
    }

    /**
     * Returns: The blocking time.
     */
    override protected @property inout(Duration) blockTime()
    inout @safe pure nothrow
    {
        return min(super.blockTime, 1.dur!"seconds");
    }
}
