/*
Copyright (c) 2016-2017 Eugene Wissner

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/**
 * Copyright: Eugene Wissner 2016-2017.
 * License: $(LINK2 boost.org/LICENSE_1_0.txt, Boost License 1.0).
 * Authors: Eugene Wissner
 */
module dlib.async.event.selector;

version (Posix):

import dlib.async.loop;
import dlib.async.transport;
import dlib.async.watcher;
import dlib.container.buffer;
import dlib.memory;
import dlib.memory.mmappool;
import dlib.network.socket;
import core.sys.posix.netinet.in_;
import core.stdc.errno;

/**
 * Transport for stream sockets.
 */
class SelectorStreamTransport : StreamTransport
{
    private ConnectedSocket socket_;

    /// Input buffer.
    package WriteBuffer input;

    private SelectorLoop loop;

    /// Received notification that the underlying socket is write-ready.
    package bool writeReady;

    /**
     * Params:
     *     loop     = Event loop.
     *     socket   = Socket.
     */
    this(SelectorLoop loop, ConnectedSocket socket)
    {
        socket_ = socket;
        this.loop = loop;
        input = MmapPool.instance.make!WriteBuffer();
    }

    /**
     * Close the transport and deallocate the data buffers.
     */
    ~this()
    {
        MmapPool.instance.dispose(input);
    }

    /**
     * Returns: Transport socket.
     */
    inout(ConnectedSocket) socket() inout pure nothrow @safe @nogc
    {
        return socket_;
    }

    /**
     * Write some data to the transport.
     *
     * Params:
     *     data = Data to send.
     */
    void write(ubyte[] data)
    {
        if (!data.length)
        {
            return;
        }
        // Try to write if the socket is write ready.
        if (writeReady)
        {
            ptrdiff_t sent;
            SocketException exception;
            try
            {
                sent = socket.send(data);
                if (sent == 0)
                {
                    writeReady = false;
                }
            }
            catch (SocketException e)
            {
                writeReady = false;
                exception = e;
            }
            if (sent < data.length)
            {
                input ~= data[sent..$];
                loop.feed(this, exception);
            }
        }
        else
        {
            input ~= data;
        }
    }
}

abstract class SelectorLoop : Loop
{
    /// Pending connections.
    protected ConnectionWatcher[] connections;

    this()
    {
        super();
        connections = MmapPool.instance.makeArray!ConnectionWatcher(maxEvents);
    }

    ~this()
    {
        foreach (ref connection; connections)
        {
            // We want to free only IOWatchers. ConnectionWatcher are created by the
            // user and should be freed by himself.
            auto io = cast(IOWatcher) connection;
            if (io !is null)
            {
                MmapPool.instance.dispose(io);
                connection = null;
            }
        }
        MmapPool.instance.dispose(connections);
    }

    /**
     * If the transport couldn't send the data, the further sending should
     * be handled by the event loop.
     *
     * Params:
     *     transport = Transport.
     *     exception = Exception thrown on sending.
     *
     * Returns: $(D_KEYWORD true) if the operation could be successfully
     *          completed or scheduled, $(D_KEYWORD false) otherwise (the
     *          transport will be destroyed then).
     */
    protected bool feed(SelectorStreamTransport transport, SocketException exception = null)
    {
        while (transport.input.length && transport.writeReady)
        {
            try
            {
                ptrdiff_t sent = transport.socket.send(transport.input[]);
                if (sent == 0)
                {
                    transport.writeReady = false;
                }
                else
                {
                    transport.input += sent;
                }
            }
            catch (SocketException e)
            {
                exception = e;
                transport.writeReady = false;
            }
        }
        if (exception !is null)
        {
            auto watcher = cast(IOWatcher) connections[transport.socket.handle];
            assert(watcher !is null);

            kill(watcher, exception);
            return false;
        }
        return true;
    }

    /**
     * Start watching.
     *
     * Params:
     *     watcher = Watcher.
     */
    override void start(ConnectionWatcher watcher)
    {
        if (watcher.active)
        {
            return;
        }

        if (connections.length <= watcher.socket)
        {
            MmapPool.instance.resizeArray(connections, watcher.socket.handle + maxEvents / 2);
        }
        connections[watcher.socket.handle] = watcher;

        super.start(watcher);
    }

    /**
     * Accept incoming connections.
     *
     * Params:
     *     connection = Connection watcher ready to accept.
     */
    package void acceptConnections(ConnectionWatcher connection)
    in
    {
        assert(connection !is null);
    }
    body
    {
        while (true)
        {
            ConnectedSocket client;
            try
            {
                client = (cast(StreamSocket) connection.socket).accept();
            }
            catch (SocketException e)
            {
                defaultAllocator.dispose(e);
                break;
            }
            if (client is null)
            {
                break;
            }

            IOWatcher io;
            auto transport = MmapPool.instance.make!SelectorStreamTransport(this, client);

            if (connections.length > client.handle)
            {
                io = cast(IOWatcher) connections[client.handle];
            }
            else
            {
                MmapPool.instance.resizeArray(connections, client.handle + maxEvents / 2);
            }
            if (io is null)
            {
                io = MmapPool.instance.make!IOWatcher(transport,
                                                      connection.protocol);
                connections[client.handle] = io;
            }
            else
            {
                io(transport, connection.protocol);
            }

            reify(io, EventMask(Event.none), EventMask(Event.read, Event.write));
            connection.incoming.insertBack(io);
        }

        if (!connection.incoming.empty)
        {
            swapPendings.insertBack(connection);
        }
    }
}
