/*
Copyright (c) 2013-2017 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

module dlib.image.compleximage;

private
{
    import dlib.math.vector;
    import dlib.math.complex;
    import dlib.math.fft;
    import dlib.image.image;
    import dlib.image.color;
    import dlib.image.signal2d;
}

class ComplexImageRGB
{
    Signal2D chR;
    Signal2D chG;
    Signal2D chB;

    this(Signal2D r, Signal2D g, Signal2D b)
    {
        chR = r;
        chG = g;
        chB = b;
    }

    this(SuperImage img)
    {
        assert(img.channels >= 3);
        chR = new Signal2D(img, 0);
        chG = new Signal2D(img, 1);
        chB = new Signal2D(img, 2);

        chR.fft();
        chG.fft();
        chB.fft();
    }

    ComplexImageRGB convolve(Signal2D kernel)
    {
        assert(kernel.domain == Signal2D.Domain.Frequency);
        return multiply(kernel);
    }

    ComplexImageRGB deconvolve(Signal2D kernel)
    {
        return divide(kernel);
    }

    ComplexImageRGB applyFilter(Signal2D filter)
    {
        assert(filter.domain == Signal2D.Domain.Spatial);
        return multiply(filter);
    }

    ComplexImageRGB multiply(Signal2D sig)
    {
        auto resR = chR.multiply(sig);
        auto resG = chG.multiply(sig);
        auto resB = chB.multiply(sig);
        return new ComplexImageRGB(resR, resG, resB);
    }

    ComplexImageRGB divide(Signal2D sig)
    {
        auto resR = chR.divide(sig);
        auto resG = chG.divide(sig);
        auto resB = chB.divide(sig);
        return new ComplexImageRGB(resR, resG, resB);
    }

    @property ImageRGB8 image()
    {
        if (chR.domain == Signal2D.Domain.Frequency)
            chR.fftInverse();
        if (chG.domain == Signal2D.Domain.Frequency)
            chG.fftInverse();
        if (chB.domain == Signal2D.Domain.Frequency)
            chB.fftInverse();
        return composeRGB(chR, chG, chB);
    }
}

ImageRGB8 composeRGB(
    Signal2D chRed,
    Signal2D chGreen,
    Signal2D chBlue)
{
    ImageRGB8 img = new ImageRGB8(chRed.width, chRed.height);

    float maxIntensityR = 0.0f;
    float maxIntensityG = 0.0f;
    float maxIntensityB = 0.0f;
    Vector3f[] fpImage = new Vector3f[img.width * img.height];

    for(int i = 0; i < chRed.data.length; i++)
    {
        float mr, mg, mb;

        mr = chRed.data[i].magnitude;
        if (mr > maxIntensityR)
            maxIntensityR = mr;
        mg = chGreen.data[i].magnitude;
        if (mg > maxIntensityG)
            maxIntensityG = mg;
        mb = chBlue.data[i].magnitude;
        if (mb > maxIntensityB)
            maxIntensityB = mb;

        fpImage[i] = Vector3f(mr, mg, mb);
    }

    foreach(ref v; fpImage)
    {
        v.r /= maxIntensityR;
        v.g /= maxIntensityG;
        v.b /= maxIntensityB;
    }

    foreach(y; 0..img.height)
    foreach(x; 0..img.width)
    {
        size_t index = (y * img.width + x);
        Vector3f m = fpImage[index];
        img[x, y] = Color4f(m);
    }

    return img;
}
